# -*- coding: utf-8 -*-
# Standard library imports
import os
import ssl
from json import dumps as json_dump
from json import loads as json_load

try:
    from urllib2 import urlopen, HTTPError, URLError
    from httplib import HTTPException
except ImportError:
    from urllib.request import urlopen
    from urllib.error import HTTPError, URLError
    from http.client import HTTPException

try:
    import certifi
except ImportError:
    from .vendor import certifi



def get(url):
    '''Make a get request.'''

    context = ssl.create_default_context(cafile=ca_certs())
    response = urlopen(url, context=context)
    return response


def json(response):
    '''Get dict from json response.'''

    return json_load(response.read().decode())


def ca_certs():
    '''Returns path to vendored certifi/cacert.pem.'''

    package = os.path.dirname(__file__)
    return os.path.join(package, 'vendor', 'certifi', 'cacert.pem')



import os
import sys

from cpenv import api, paths, shell
from cpenv.cli import core
from cpenv.resolver import ResolveError


class Clone(core.CLI):
    '''
    Clone a Module for development.

    The following repos are available by default:
        home - Local repo pointing to a computer-wide cpenv directory
        user - Local repo pointing to a user-specific cpenv directory

    For a full listing of available repos use the repo cli command:
        cpenv repo list
    '''

    def setup_parser(self, parser):
        parser.add_argument(
            'module',
            help='Module to clone.',
        )
        parser.add_argument(
            'where',
            help='Destination directory. (./<module_name>)',
            nargs='?',
            default=None,
        )
        parser.add_argument(
            '--from_repo',
            help='Specific repo to clone from.',
            default=None,
        )
        parser.add_argument(
            '--overwrite',
            help='Overwrite the destination directory. (False)',
            action='store_true',
        )

    def run(self, args):

        try:
            if not args.from_repo:
                module_spec = api.resolve([args.module])[0]
            else:
                from_repo = api.get_repo(args.from_repo)
                module_spec = from_repo.find(args.module)[0]
        except Exception:
            core.echo()
            core.echo('Error: Failed to resolve ' + args.module)
            core.exit(1)

        where = paths.normalize(args.where or '.', module_spec.real_name)
        if os.path.isdir(where):
            core.echo('Error: Directory already exists - ' + where)
            core.exit(1)

        core.echo('- Cloning %s...' % module_spec.real_name)
        core.echo()
        try:
            module = module_spec.repo.download(
                module_spec,
                where=where,
                overwrite=args.overwrite,
            )
        except Exception as e:
            core.echo()
            core.echo('Error: ' + str(e))
            core.exit(1)

        core.echo()
        core.echo('Navigate to the following folder to make changes:')
        core.echo('  ' + module.path)
        core.echo()
        core.echo("Use one of the following commands to publish your changes:")
        core.echo('  cpenv publish .')
        core.echo('  cpenv publish . --to_repo="repo_name"')



from cpenv import api
from cpenv.cli import core
from cpenv.module import is_partial_match, sort_modules


class List(core.CLI):
    '''List active and available Modules.'''

    def setup_parser(self, parser):
        parser.add_argument(
            'requirement',
            help='Space separated list of modules.',
            nargs='?',
            default=None,
        )
        parser.add_argument(
            '--repo',
            help='List modules only in this repo.',
            default=None,
        )
        parser.add_argument(
            '--verbose', '-v',
            help='Print more module info.',
            action='store_true',
        )

    def run(self, args):

        found_modules = False

        core.echo()
        active_modules = api.get_active_modules()
        if args.requirement:
            active_modules = [
                m for m in active_modules
                if is_partial_match(args.requirement, m)
            ]

        if active_modules:
            found_modules = True
            core.echo(core.format_columns(
                '[*] Active',
                [m.real_name for m in sort_modules(active_modules)],
            ))
            core.echo()

        repos = api.get_repos()
        if args.repo:
            repos = [api.get_repo(args.repo)]

        for repo in repos:
            if args.requirement:
                repo_modules = repo.find(args.requirement)
            else:
                repo_modules = repo.list()

            module_names = []
            for module in sort_modules(repo_modules):
                if module in active_modules:
                    module_names.append('* ' + module.real_name)
                else:
                    module_names.append('  ' + module.real_name)

            if module_names:
                found_modules = True
                if repo.name != repo.path:
                    header = repo.name + ' - ' + repo.path
                else:
                    header = repo.name
                core.echo(core.format_columns(
                    '[ ] ' + header,
                    module_names,
                    indent='  ',
                ))
                core.echo()

        if not found_modules:
            core.echo('No modules are available.')



import cpenv
from cpenv import api, paths
from cpenv.cli import core


class Version(core.CLI):
    '''Show version information.'''

    def run(self, args):

        core.echo()
        core.echo(core.format_section(
            'Version Info',
            [
                ('name', cpenv.__name__),
                ('version', cpenv.__version__),
                ('url', cpenv.__url__),
                ('package', paths.parent(cpenv.__file__)),
                ('path', api.get_module_paths()),
            ]
        ))
        core.echo()

        # List package versions
        dependencies = []
        try:
            import Qt
            dependencies.extend([
                ('Qt.py', Qt.__version__),
                ('Qt Binding', Qt.__binding__ + '-' + Qt.__binding_version__),
            ])
        except ImportError:
            pass

        if not dependencies:
            return

        core.echo(core.format_section('Dependencies', dependencies), end='\n\n')



import sys

from cpenv import api, reporter
from cpenv.cli import core
from cpenv.resolver import ResolveError


class Info(core.CLI):
    '''Show Module info'''

    usage = 'cpenv info [-h] [<modules>...]'

    def setup_parser(self, parser):
        parser.add_argument(
            'modules',
            help='Space separated list of modules.',
            nargs='*',
        )
        parser.add_argument(
            '--key',
            help=(
                'Specify the name of a key like "path" or "version". '
                'When provided this command will print the value of the key.'
            ),
            default=None,
        )
        parser.add_argument(
            '--repo',
            help='Get repo info instead of module info.',
            action='store_true',
        )
        parser.add_argument(
            '--home',
            help='Get the path to the cpenv home directory.',
            action='store_true',
        )
        parser.add_argument(
            '--cache',
            help='Get the path to the cpenv cache directory.',
            action='store_true',
        )

    def _get_key(self, obj, key):
        try:
            return eval('obj.' + key)
        except Exception as e:
            core.echo('Error: %s' % e)
            sys.exit(1)

    def run(self, args):
        if args.home:
            core.echo(api.get_home_path())
            return

        if args.cache:
            core.echo(api.get_cache_path())
            return

        # Key mode - print a key's value of a module or repo.
        if args.key:
            reporter.set_reporter(reporter.Reporter())
            if args.repo:
                objects = []
                for repo_name in args.modules:
                    repo = api.get_repo(repo_name)
                    if repo is None:
                        core.echo('Could not find repo %s' % repo_name)
                        sys.exit(1)
                    objects.append(repo)
            else:
                try:
                    objects = api.resolve(args.modules)
                except ResolveError:
                    sys.exit(1)

            value = self._get_key(objects[0], args.key)
            core.echo(value)
            return

        # Normal mode - resolve repos or modules and show info.
        core.echo()
        if args.repo:
            repos = []
            for repo_name in args.modules:
                repo = api.get_repo(repo_name)
                if repo is None:
                    core.echo('Could not find repo %s' % repo_name)
                    sys.exit(1)
                repos.append(repo)

            for repo in repos:
                core.echo(core.format_section(
                    repo.name,
                    [
                        ('name', repo.name),
                        ('path', repo.path),
                        ('type', str(type(repo))),
                        ('type_name', repo.type_name),
                    ]
                ))
                core.echo()
            return
        else:
            try:
                module_specs = api.resolve(args.modules)
            except ResolveError:
                sys.exit(1)

            core.echo()
            for spec in module_specs:
                core.echo(core.format_section(
                    spec.qual_name,
                    [
                        (k, str(v))
                        for k, v in sorted(spec._asdict().items())
                    ],
                ))
                core.echo()
            return



from cpenv import api
from cpenv.cli import core
from cpenv.module import Module
from cpenv.resolver import Resolver


class Publish(core.CLI):
    '''Publish a Module to a repo.'''

    def setup_parser(self, parser):
        parser.add_argument(
            'module',
            help='Path of module to publish. (".")',
            default='.',
            nargs='?',
        )
        parser.add_argument(
            '--to_repo',
            help='Specific repo to clone from.',
            default=None,
        )
        parser.add_argument(
            '--overwrite',
            help='Overwrite the destination directory. (False)',
            action='store_true',
        )

    def run(self, args):

        core.echo()

        # Get repo
        if args.to_repo:
            to_repo = api.get_repo(name=args.to_repo)
        else:
            to_repo = core.prompt_for_repo(
                api.get_repos(),
                'Choose a repo to publish to',
                default_repo_name='home',
            )

        # Resolve module
        resolver = Resolver(api.get_repos())
        module_spec = resolver.resolve([args.module])[0]
        core.echo()

        # Confirm publication
        choice = core.prompt('Publish module to %s?[y/n] ' % to_repo.name)
        if choice.lower() not in ['y', 'yes', 'yup']:
            core.echo('Aborted.')
            sys.exit(1)

        # Publish
        module = Module(module_spec.path)
        published = to_repo.upload(module, args.overwrite)
        core.echo()

        core.echo('Activate your module:')
        core.echo('  cpenv activate %s' % published.real_name)
        core.echo()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function

missing = object()


class Environment(object):
    '''An object that represents a list of requirements.

    Once you have activated some modules, you can save them as an Environment
    to be activated later. Think of them as an alias for activating a list
    of module requirements.

    # TODO: Add the ability to save and activate Environments directly from
    all Repos. The Shotgun app tk-cpenv implements this functionality
    internally, it should be moved to cpenv at some point and the code related
    to Environments in tk-cpenv can be removed.
    '''

    def __init__(self, name, data=None, path=None):
        self.name = name
        self.data = data or {'name': name, 'requires': []}
        self.path = path

    def get(self, key, default=missing):
        if key not in self.data and default is not missing:
            return default

        return self.data.get(key)

    def set(self, key, value):
        self.data[key] = value

    @property
    def requires(self):
        return self.data['requires']



import argparse
import re

from cpenv import api, repos
from cpenv.cli import core


class Repo(core.CLI):
    '''
    Manage, Configure and list Repos.

    Repos are sources of Modules. A Repo can be local or remote. Modules
    stored in remote Repos will be localized to a local repo prior to
    activation or explicitly by using "cpenv localize".
    '''

    def commands(self):
        return [
            ListRepo(self),
            AddRepo(self),
            RemoveRepo(self),
        ]


class ListRepo(core.CLI):
    '''List configured repos'''

    name = 'list'

    def run(self, args):

        core.echo()
        core.echo('Repos in order of resolution:')
        for i, repo in enumerate(api.get_repos()):
            if repo.path == repo.name:
                core.echo('  [{}] {}  {}'.format(
                    i,
                    type(repo).__name__,
                    repo.path,
                ))
            else:
                core.echo('  [{}] {}  {}  {}'.format(
                    i,
                    type(repo).__name__,
                    repo.name,
                    repo.path,
                ))


class AddRepo(core.CLI):
    '''Add a new repo.

    LocalRepo:
        cpenv repo add custom --path=~/custom_repo

    ShotgunRepo:
        cpenv repo add my_shotgun --type=shotgun --base_url=https://my.shotgunstudio.com --script_name=cpenv --api_key=secret
    '''

    name = 'add'

    def setup_parser(self, parser):
        parser.add_argument(
            '--type',
            help='Type of repo',
            choices=['local', 'shotgun'],
            default='local',
        )
        parser.add_argument(
            'name',
            help='Name of the repo',
        )
        parser.add_argument(
            'type_args',
            help='Type specific arguments.',
            nargs=argparse.REMAINDER,
        )

    def parse_type_args(self, type_args):
        pattern = (
            r'-{1,2}(?P<param>[a-zA-Z0-9_]+)=*\s*'
            r'"?(?P<value>.+)"?'
        )
        kwargs = {}
        for arg in type_args:
            match = re.search(pattern, arg)
            param = match.group('param')
            raw_value = '"%s"' % match.group('value').strip()
            value = core.safe_eval(raw_value)
            kwargs[param] = value
        return kwargs

    def run(self, args):

        # Parse type_args or args that are specific to a given Repo type
        repo_kwargs = self.parse_type_args(args.type_args)
        repo_type = repo_kwargs.pop('type', args.type)

        core.echo()
        if repo_type not in repos.registry:
            core.echo('Error: %s is not a registered repo type.' % args.type)
            core.echo('Choose from: ' + ', '.join(repos.registry.keys()))
            core.exit(1)

        if api.get_repo(args.name):
            core.echo('Error: Repo named %s already exists.' % args.name)
            core.exit(1)

        repo_cls = repos.registry[repo_type]
        repo_kwargs['name'] = args.name
        core.echo('- Checking %s args...' % repo_cls.__name__, end='')
        try:
            repo_cls(**repo_kwargs)
        except Exception as e:
            core.echo('OOPS!')
            core.echo()
            core.echo('Error: Failed to initialize %s' % repo_cls.__name__)
            core.echo(str(e))
            core.exit(1)
        core.echo('OK!')
        core.echo()

        core.echo('- Adding repo to config...', end='')
        repo_kwargs['type'] = repo_type
        repo_config = api.read_config('repos', {})
        repo_config[args.name] = repo_kwargs
        api.write_config('repos', repo_config)
        core.echo('OK!')
        core.echo()


class RemoveRepo(core.CLI):
    '''Remove repo by name.'''

    name = 'remove'

    def setup_parser(self, parser):
        parser.add_argument(
            'name',
            help='Name of the repo',
        )

    def run(self, args):
        core.echo()
        if args.name in ['home', 'user', 'cwd']:
            core.echo('Error: Can not remove %s repo.' % args.name)

        repo_config = api.read_config('repos', {})
        if args.name not in repo_config:
            core.echo('Error: Repo named %s not found.' % args.name)
            core.exit(1)

        core.echo('- Removing repo from config...', end='')
        repo_config.pop(args.name)
        api.write_config('repos', repo_config)
        core.echo('OK!')
        core.echo()



from cpenv import api
from cpenv.cli import core
from cpenv.resolver import Copier, Resolver


class Copy(core.CLI):
    '''Copy Modules from one Repo to another.'''

    def setup_parser(self, parser):
        parser.add_argument(
            'modules',
            help='Space separated list of modules.',
            nargs='+',
        )
        parser.add_argument(
            '--from_repo',
            help='Download from',
            default=None,
        )
        parser.add_argument(
            '--to_repo',
            help='Upload to',
            default=None,
        )
        parser.add_argument(
            '--overwrite',
            help='Overwrite the destination directory. (False)',
            action='store_true',
        )

    def run(self, args):

        core.echo()

        if args.from_repo:
            from_repo = api.get_repo(args.from_repo)
        else:
            from_repo = core.prompt_for_repo(
                api.get_repos(),
                'Download from',
                default_repo_name='home',
            )

        if args.to_repo:
            to_repo = api.get_repo(args.to_repo)
        else:
            repos = api.get_repos()
            repos.remove(from_repo)
            to_repo = core.prompt_for_repo(
                repos,
                'Upload to',
                default_repo_name='home',
            )

        resolver = Resolver([from_repo])
        module_specs = resolver.resolve(args.modules)
        core.echo()

        choice = core.prompt('Copy these modules to %s?[y/n] ' % to_repo.name)
        if choice.lower() not in ['y', 'yes', 'yup']:
            core.echo('Aborted.')
            core.exit(1)
        core.echo()

        copier = Copier(to_repo)
        copier.copy(module_specs, args.overwrite)



import atexit
import sys

from cpenv import api, cli, reporter
from cpenv.cli import core


def perform_self_version_check():
    from cpenv import _self_version_check
    is_latest, current, latest = _self_version_check.is_latest_version()
    if not is_latest:
        _self_version_check.warn_newer_version_available(current, latest)


def main():
    atexit.register(perform_self_version_check)
    reporter.set_reporter(cli.CliReporter())
    core.run(cli.CpenvCLI, sys.argv[1:])


if __name__ == '__main__':
    main()



import os
import sys

from cpenv import api, repos, shell
from cpenv.cli import core
from cpenv.resolver import ResolveError


class Edit(core.CLI):
    '''Open a module in a text editor.

    Editor Lookup:
        1. CPENV_EDITOR environment variable
        2. EDITOR environment variable
        3. subl (default editor)
    '''

    usage = 'cpenv edit [-h, --env] <module_or_environment>'

    def setup_parser(self, parser):
        parser.add_argument(
            '--env',
            help='Edit an Environment instead of a Module.',
            action='store_true',
        )
        parser.add_argument(
            'module_or_environment',
            help='Module or Environment name.',
        )

    def run(self, args):
        core.echo()
        if args.env:
            return self.run_env(args.module_or_environment)

        try:
            module_spec = api.resolve([args.module_or_environment])[0]
        except ResolveError:
            sys.exit(1)

        if not isinstance(module_spec.repo, repos.LocalRepo):
            core.echo('%s - %s' % (module_spec.qual_name, module_spec.path))
            core.echo('Error: Can only edit modules in local repositories.')
            sys.exit(1)

        editor = os.getenv('CPENV_EDITOR', os.getenv('EDITOR', 'subl'))
        core.echo('Opening %s in %s.' % (module_spec.path, editor))
        shell.run(editor, module_spec.path)

    def run_env(self, environment):
        file = None
        repo = False
        found_in_remote = False
        for repo in api.get_repos():
            for env in repo.list_environments():
                if env.name == environment:
                    if isinstance(repo, repos.LocalRepo):
                        file = env.path
                        break
                    else:
                        repo = repo
                        found_in_remote = True
            if file:
                break

        if found_in_remote:
            core.echo('Error: Can only edit Environments in local repos.')
            core.echo('Found %s in repo %s' % (environment, repo.name))
            sys.exit(1)

        editor = os.getenv('CPENV_EDITOR', os.getenv('EDITOR', 'subl'))
        core.echo('Opening %s in %s.' % (file, editor))
        shell.run(editor, file)



import tqdm

from cpenv import repos
from cpenv.cli import (
    activate,
    clone,
    copy,
    core,
    create,
    edit,
    env,
    info,
    list,
    localize,
    publish,
    remove,
    repo,
    version,
)
from cpenv.reporter import Reporter


class CpenvCLI(core.CLI):
    '''
    Create, activate and manage Modules.

    A Module is a folder containing a module.yml file describing the Module's
    name, version, requirements, and environment variables. Module's can also
    contain a hooks folder with python files that respond to specific events.
    '''

    name = 'cpenv'
    usage = 'cpenv [-h] <command> [<args>...]'

    def commands(self):
        return [
            activate.Activate(self),
            clone.Clone(self),
            copy.Copy(self),
            create.Create(self),
            info.Info(self),
            edit.Edit(self),
            env.Env(self),
            list.List(self),
            localize.Localize(self),
            publish.Publish(self),
            remove.Remove(self),
            repo.Repo(self),
            version.Version(self),
        ]


class CliReporter(Reporter):

    def __init__(self):
        self._bars = {}

    def get_bar_style(self, desc, total, unit=None):
        return {
            'desc': desc,
            'total': total,
            'bar_format': '  {desc} {bar} {n_fmt}/{total_fmt}',
            'unit': unit or 'iB',
            'unit_scale': True,
        }

    def start_resolve(self, requirements):
        core.echo('- Resolving requirements...')
        core.echo()

    def resolve_requirement(self, requirement, module_spec):
        core.echo('  %s - %s' % (module_spec.real_name, module_spec.path))

    def end_resolve(self, resolved, unresolved):
        core.echo()
        if unresolved:
            core.echo('Error: Failed to resolve %s' % ', '.join(unresolved))

    def start_localize(self, module_specs):
        for spec in module_specs:
            if not isinstance(spec.repo, repos.LocalRepo):
                core.echo('- Localizing modules...')
                core.echo()
                return

    def end_localize(self, modules):
        core.echo()

    def start_progress(self, label, max_size, data):
        if 'download' in label.lower():
            spec = data['module_spec']
            core.echo(
                '  Downloading %s from %s...' %
                (spec.qual_name, spec.repo.name)
            )
            desc = spec.real_name
        elif 'upload' in label.lower():
            module = data['module']
            to_repo = data['to_repo']
            core.echo(
                '  Uploading %s to %s...' %
                (module.qual_name, to_repo.name)
            )
            desc = module.real_name
        else:
            desc = label

        style = self.get_bar_style(desc, max_size, data.get('unit', None))
        self._bars[label] = tqdm.tqdm(**style)

    def update_progress(self, label, chunk_size, data):
        bar = self._bars.get(label, None)
        if bar:
            bar.update(chunk_size)

    def end_progress(self, label, data):
        bar = self._bars.pop(label, None)
        if bar:
            bar.close()



# -*- coding: utf-8 -*-
'''
Tools for working with paths and files.
'''

# Standard library imports
import os
import shutil
import stat
import zipfile
from fnmatch import fnmatch


def normalize(*parts):
    '''Join, expand, and normalize a filepath.'''

    path = os.path.expanduser(os.path.expandvars(os.path.join(*parts)))
    return os.path.abspath(path).replace('\\', '/')


def parent(path):
    '''Get a file paths parent directory'''
    return normalize(os.path.dirname(path))


def ensure_path_exists(path, *args):
    '''Like os.makedirs but keeps quiet if path already exists'''

    if os.path.exists(path):
        return

    os.makedirs(path, *args)


def is_writable(path):
    '''Check if a directory is writable.'''

    if os.path.exists(path):
        return os.access(path, os.X_OK | os.W_OK)

    try:
        os.makedirs(path)
        touch(normalize(path, 'tmpfile'))
    except OSError as e:
        return False
    else:
        rmtree(path)
        return True


def rmtree(path):
    '''Safely remove directory and all of it's contents.'''

    def onerror(func, path, _):
        os.chmod(path, stat.S_IWRITE)
        func(path)

    shutil.rmtree(path, onerror=onerror)


def walk_up(start_dir, depth=20):
    '''Like os.walk but walks up a tree.'''

    root = start_dir

    for i in range(depth):
        contents = os.listdir(root)
        subdirs, files = [], []
        for f in contents:
            if os.path.isdir(os.path.join(root, f)):
                subdirs.append(f)
            else:
                files.append(f)

        yield root, subdirs, files

        parent = os.path.dirname(root)
        if parent and not parent == root:
            root = parent
        else:
            break


def touch(filepath):
    '''Touch the given filepath'''

    with open(filepath, 'a'):
        os.utime(filepath, None)


def format_size(bytesize):
    '''Human readable size.'''

    value = bytesize
    for unit in ['b', 'kb', 'mb', 'gb', 'tb', 'pb', 'zi']:
        if value < 1024.0:
            return "{:.0f} {}".format(value, unit)
        value /= 1024.0
    return "{:.0f} {}".format(value, unit)


def get_file_count(folder):
    '''Get the number of files in a folder.'''

    count = 0
    for _, _, files in exclusive_walk(folder):
        count += len(files)
    return count


def get_folder_size(folder):
    '''Get the size of a folder in bytes.'''

    size = 0
    for root, _, files in exclusive_walk(folder):
        for file in files:
            file_path = os.path.join(root, file)
            if not os.path.islink(file_path):
                size += os.path.getsize(file_path)
    return size


def exclude_names(names):
    '''Returns True when a file matches one of the provided names.'''

    def check_file_against_names(file):
        return os.path.basename(file) in names

    return check_file_against_names


def exclude_patterns(patterns):
    '''Returns True when a file matches against one of the provided patterns'''

    def check_file_against_patterns(file):
        return any([fnmatch(file, p) for p in patterns])

    return check_file_against_patterns


def include_prebuilt_pyc(file):
    '''Returns True when a pyc file has no accompanying py file.

    This is useful in cases where tool distributors prebuild pyc files as
    a crude form of anti-piracy. These need to be included in cpenv modules
    when zipped to be uploaded to a repository.
    '''

    return (
        file.endswith('.pyc')
        and not os.path.isfile(file.replace('.pyc', '.py'))
        and '__pycache__' not in file
    )


def is_excluded(value, predicates):
    '''Check if a value matches any of the exclude predicates.'''

    return any([predicate(value) for predicate in predicates])


def is_included(value, predicates):
    '''Returns True if value matches any of the include predicates.'''

    return any([predicate(value) for predicate in predicates])


def exclusive_walk(folder, excludes=None, includes=None):
    '''Like os.walk but excludes/includes files by using predicate functions.

    Excludes the following by default:
        names: __pycache__, .git, Thumbs.db, .venv, venv
        patterns: *.pyc, *.egg-info

    Includes the following by default:
        .pyc files that have no accompanying .py file.

    Arguments:
        folder (str): Root folder to recursively walk.
        excludes ([callable]): List of predicate functions used to exclude files.
        includes ([callable]): List of predicate functions to include files. Overrides excludes.

    Returns:
        Generator yielding (root, subdirs, files).
    '''

    excludes = excludes or [
        exclude_names(['__pycache__', '.git', 'thumbs.db', '.venv', 'venv']),
        exclude_patterns(['*.pyc', '*.egg-info']),
    ]
    includes = includes or [include_prebuilt_pyc]

    for root, subdirs, files in os.walk(folder):
        if is_excluded(root, excludes) and not is_included(root, includes):
            subdirs[:] = []
            continue

        included_files = []
        for file in files:
            path = normalize(root, file)
            if is_excluded(path, excludes) and not is_included(path, includes):
                continue
            included_files.append(file)

        yield root, subdirs, included_files


def zip_folder(folder, where):
    '''Zip the contents of a folder.'''

    parent = os.path.dirname(where)
    if not os.path.isdir(parent):
        os.makedirs(parent)

    # TODO: Count files first so we can report progress of building zip.

    with zipfile.ZipFile(where, 'w', zipfile.ZIP_DEFLATED) as zip_file:
        for root, subdirs, files in exclusive_walk(folder):
            rel_root = os.path.relpath(root, folder)
            for file in files:
                zip_file.write(
                    os.path.join(root, file),
                    arcname=os.path.join(rel_root, file)
                )



from cpenv import api
from cpenv.cli import core
from cpenv.module import best_match


class Remove(core.CLI):
    '''Permanently remove a Module from a Repo.'''

    def setup_parser(self, parser):
        parser.add_argument(
            'modules',
            help='Space separated list of modules.',
            nargs='+',
        )
        parser.add_argument(
            '--from_repo',
            help='Specific repo to remove from.',
            default=None,
        )

    def run(self, args):

        core.echo()

        if args.from_repo:
            from_repo = api.get_repo(name=args.from_repo)
        else:
            from_repo = core.prompt_for_repo(
                api.get_repos(),
                'Choose a repo to remove module from',
                default_repo_name='home',
            )

        core.echo('- Finding modules in %s...' % from_repo.name)
        core.echo()
        modules_to_remove = []
        for requirement in args.modules:
            module = best_match(requirement, from_repo.find(requirement))
            if not module:
                core.echo('Error: %s not found...' % requirement)
                core.exit(1)
            core.echo('  %s - %s' % (module.real_name, module.path))
            modules_to_remove.append(module)

        core.echo()
        choice = core.prompt('Delete these modules?[y/n] ')
        if choice.lower() not in ['y', 'yes', 'yup']:
            core.echo('Aborted.')
            core.exit(1)

        core.echo()
        core.echo('- Removing modules...')
        core.echo()
        for module in modules_to_remove:
            core.echo('  ' + module.real_name)
            api.remove(module, from_repo)

        core.echo()
        core.echo('Successfully removed modules.')
        core.echo()



from cpenv import api
from cpenv.cli import core
from cpenv.repos import LocalRepo
from cpenv.resolver import Localizer, ResolveError


class Localize(core.CLI):
    '''Localize a list of Modules.

    Downloads modules from a remote Repo and places them in the home LocalRepo
    by default. Use the --to_repo option to specify a LocalRepo.
    '''

    def setup_parser(self, parser):
        parser.add_argument(
            'modules',
            help='Space separated list of modules.',
            nargs='+',
        )
        parser.add_argument(
            '--to_repo', '-r',
            help='Specific repo to localize to. (first match)',
            default=None,
        )
        parser.add_argument(
            '--overwrite', '-o',
            help='Overwrite the destination directory. (False)',
            action='store_true',
        )

    def run(self, args):

        core.echo()

        if args.to_repo:
            to_repo = api.get_repo(name=args.to_repo)
        else:
            repos = [r for r in api.get_repos() if isinstance(r, LocalRepo)]
            to_repo = core.prompt_for_repo(
                repos,
                'Choose a repo to localize to',
                default_repo_name='home',
            )

        # Resolve modules
        try:
            resolved = api.resolve(args.modules)
        except ResolveError:
            core.exit(1)

        # Localize modules from remote repos
        remote_modules = [
            spec for spec in resolved
            if not isinstance(spec.repo, LocalRepo)
        ]

        if not len(remote_modules):
            core.echo('All modules are already local...')
            core.exit()

        try:
            localizer = Localizer(to_repo)
            localizer.localize(
                remote_modules,
                overwrite=args.overwrite,
            )
        except Exception as e:
            core.echo('Error: ' + str(e))
            core.exit(1)



from cpenv import api, shell
from cpenv.cli import core
from cpenv.resolver import ResolveError


class Activate(core.CLI):
    '''Activate a list of Modules or an Environment.

    Examples:
      cpenv activate module_a module_b
      cpenv activate module_a-1.0 module_b 0.2.0
      cpenv activate my_environment
      cpenv activate --env my_environment

    Note:
      Use the --env flag to specifically activate an Environment by name
      rather than checking for modules first. Use the "cpenv env" command to
      manage Environments.
    '''

    usage = 'cpenv activate [-h] [<modules> or <environment>...]'

    def setup_parser(self, parser):
        parser.add_argument(
            'modules',
            help='Space separated list of modules.',
            nargs='+',
        )
        parser.add_argument(
            '--env',
            help='Activate an Environment. (False)',
            action='store_true',
        )

    def run(self, args):

        core.echo()

        if args.env:
            try:
                api.activate_environment(args.modules[0])
            except ResolveError as e:
                core.echo(str(e))
                core.exit(1)
        else:
            try:
                api.activate(args.modules)
            except ResolveError:
                api.activate_environment(args.modules[0])
            except ResolveError:
                core.exit(1)

        core.echo('- Launching subshell...')
        core.echo()
        core.echo('  Type "exit" to deactivate.')
        core.echo()
        shell.launch('[*]')



import argparse
import re

from cpenv import api
from cpenv.cli import core


class Env(core.CLI):
    '''
    Manage, Configure and list Repos.

    Repos are sources of Modules. A Repo can be local or remote. Modules
    stored in remote Repos will be localized to a local repo prior to
    activation or explicitly by using "cpenv localize".
    '''

    def commands(self):
        return [
            ListEnv(self),
            SaveEnv(self),
            RemoveEnv(self),
        ]


class ListEnv(core.CLI):
    '''
    List available Environments.

    Environments are like Aliases for a list of module requirements. They
    can be Activated using a single name rather than providing a full list of
    module requirements.
    '''

    name = 'list'

    def setup_parser(self, parser):
        parser.add_argument(
            'filters',
            help='Environment filters like --name="My*".',
            nargs=argparse.REMAINDER,
        )

    def parse_filters(self, filters):
        pattern = (
            r'-{1,2}(?P<param>[a-zA-Z0-9_]+)=*\s*'
            r'"?(?P<value>.+)"?'
        )
        kwargs = {}
        for arg in filters:
            match = re.search(pattern, arg)
            param = match.group('param')
            raw_value = '"%s"' % match.group('value').strip()
            value = core.safe_eval(raw_value)
            kwargs[param] = value
        return kwargs

    def run(self, args):

        filters = self.parse_filters(args.filters)
        core.echo()

        found_environments = False

        for repo in api.get_repos():
            environments = repo.list_environments(filters)
            env_names = sorted([env.name for env in environments])
            if env_names:
                found_environments = True
                if repo.name != repo.path:
                    header = repo.name + ' - ' + repo.path
                else:
                    header = repo.name
                core.echo(core.format_columns(
                    header,
                    env_names,
                    indent='  ',
                ))
                core.echo()

        if not found_environments:
            core.echo('No Environments are were found.')
            core.echo('Use "cpenv env save <name>" to save an Environment.')


class SaveEnv(core.CLI):
    '''
    Save an Environment.

    Examples:
        cpenv env save my_env
        cpenv env save --force my_env
        cpenv env save --no_versions my_env
        cpenv env save --repo MyShotgunRepo --project=1234 my_env
    '''

    name = 'save'

    def setup_parser(self, parser):
        parser.add_argument(
            '--repo',
            help='Repo to save the Environment in.',
            default='home',
        )
        parser.add_argument(
            '--force',
            help='Overwrite existing Environment? (False)',
            action='store_true',
        )
        parser.add_argument(
            '--no_versions',
            help='Write requirements without versions. (False)',
            action='store_true',
        )
        parser.add_argument(
            'name',
            help='Name of the Environment.'
        )
        parser.add_argument(
            'data',
            help='Data to include in Environment.',
            nargs=argparse.REMAINDER,
        )

    def parse_data(self, data):
        pattern = (
            r'-{1,2}(?P<param>[a-zA-Z0-9_]+)=*\s*'
            r'"?(?P<value>.+)"?'
        )
        kwargs = {}
        for arg in data:
            match = re.search(pattern, arg)
            param = match.group('param')
            raw_value = '"%s"' % match.group('value').strip()
            value = core.safe_eval(raw_value)
            kwargs[param] = value
        return kwargs

    def run(self, args):

        extra_data = self.parse_data(args.data)
        repo = api.get_repo(args.repo)
        requires = api.get_active_modules()

        if not requires:
            core.echo('Activate some modules before you save an Environment.')
            return

        if args.no_versions:
            requires = [m.name for m in requires]
        else:
            requires = [m.qual_name for m in requires]

        data = {
            'name': args.name,
            'requires': requires,
        }
        data.update(extra_data)

        repo.save_environment(
            name=data['name'],
            data=data,
            force=args.force,
        )


class RemoveEnv(core.CLI):
    '''
    Remove an Environment.

    Examples:
      cpenv env remove my_env
      cpenv env remove my_env --from_repo SomeRepo
    '''

    name = 'remove'

    def setup_parser(self, parser):
        parser.add_argument(
            'name',
            help='Name of the Environment.'
        )
        parser.add_argument(
            '--from_repo',
            help='Repo to remove the Environment from.',
            default='home',
        )

    def run(self, args):

        name = args.name
        from_repo = api.get_repo(name=args.from_repo)

        core.echo()
        core.echo('- Removing Environment from %s...' % from_repo.name)
        core.echo()

        env_removed = False
        for env in from_repo.list_environments():
            if env.name == name:
                env_removed = True
                from_repo.remove_environment(env.name)

        if env_removed:
            core.echo('Successfully removed %s.' % name)
        else:
            core.echo('Could not find Environment "%s".' % name)

        core.echo()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cpenv',
 'cpenv.cli',
 'cpenv.repos',
 'cpenv.vendor',
 'cpenv.vendor.cachetools',
 'cpenv.vendor.certifi',
 'cpenv.vendor.shotgun_api3',
 'cpenv.vendor.shotgun_api3.lib',
 'cpenv.vendor.shotgun_api3.lib.certifi',
 'cpenv.vendor.shotgun_api3.lib.httplib2',
 'cpenv.vendor.shotgun_api3.lib.httplib2.python2',
 'cpenv.vendor.shotgun_api3.lib.httplib2.python3',
 'cpenv.vendor.shotgun_api3.lib.mockgun',
 'cpenv.vendor.yaml',
 'cpenv.vendor.yaml.yaml2',
 'cpenv.vendor.yaml.yaml3']

package_data = \
{'': ['*'], 'cpenv': ['bin/*']}

install_requires = \
['colorama>=0.4.3,<0.5.0', 'psutil>=5.7.0,<6.0.0', 'tqdm>=4.46.0,<5.0.0']

entry_points = \
{'console_scripts': ['cpenv = cpenv.__main__:main']}

setup_kwargs = {
    'name': 'cpenv',
    'version': '0.5.28',
    'description': 'Cross-platform module and environment management.',
    'long_description': '<p align="center">\n    <img src="https://raw.github.com/cpenv/cpenv/master/res/icon_dark.png"/>\n</p>\n\n# cpenv\nManage software plugins, project dependencies and environment\nvariables using Modules.\n\n<p align="center">\n    <img src="https://raw.github.com/cpenv/cpenv/master/res/demo.gif"/>\n</p>\n\n# Installation\nThe recommended method of installing cpenv is via [pipx](https://pipxproject.github.io/pipx).\nPipx is used to install python cli applications in isolation.\n\n```\npipx install cpenv\n```\n```\npipx upgrade cpenv\n```\n```\npipx uninstall cpenv\n```\n\n# Overview\nCpenv is a cli tool and python library used to create, edit, publish, and activate Modules. A Module is a folder containing a dependency, like Arnold for Maya, and a module file that configures it.\n\n## Environment Variables\n| Variable                 | Description                            | Default |\n| ------------------------ | -------------------------------------- | ------- |\n| CPENV_HOME               | Customize path to cpenv home           |         |\n| CPENV_DISABLE_PROMPT     | Disable prompt when modules activated  | 0       |\n| CPENV_ACTIVE_MODULES     | List of activated modules              |         |\n| CPENV_SHELL              | Preferred subshell like "powershell"   |         |\n\n## Example Modules\n- [snack](https://github.com/cpenv/snack)\n\n## Create a Module\nUse `cpenv create <module>` to use a guide to create a new Module.\nThis will create a new folder in your current working directory with a module file in it.\n\n## Edit a Module\nEach Module contains a module.yml file, referred to as a module file. A module file contains metadata like the name and version of a module, as well as configuration, like environment variables.\n```\n# Variables\n# $MODULE - path to this module\n# $PLATFORM - platform name (win, mac, linux)\n# $PYVER - python version (2.7, 3.6...)\n\n# Wrap variables in brackets when they are nested within a string.\n#     DO \'this${variable}isnested/\' NOT \'this$variableisnested\'\n\nname: \'my_module\'\nversion: \'0.1.0\'\ndescription: \'My first module.\'\nauthor: \'Me\'\nemail: \'me@email.com\'\nrequires: []\nenvironment:\n    MY_MODULE_VAR: \'Test\'\n```\n\n### Environment key\nSetting a value will insert a key or overwrite it\'s existing value.\n```\nSOME_VAR: \'SOME_VALUE\'\n```\n\nUse the $MODULE variable for module relative paths.\n```\nRELATIVE_VAR: $MODULE/bin\n```\n\nUse lists to prepend values to a key.\n```\nPATH:\n    - $MODULE/bin\n```\n\nUse `win`, `linux` and `mac or osx` keys to declare platform specific values. If you leave out a platform, the variable will not be included on that platform.\n```\nPLATFORM_VAR:\n    mac: /mnt/some/path\n    linux: /Volumes/some/path\n    win: C:/some/path\n```\n\nYou can also use platform keys when prepending values to a variable.\n```\nPATH:\n    - mac: $MODULE/macattack\n      linux: $MODULE/penguin\n      win: $MODULE/squares\n```\n\nReuse environment variables to simplify things.\n```\nBASE: $MODULE/$PLATFORM/base\nPATH:\n    - $BASE/bin\nPYTHONPATH:\n    - $BASE/python\n```\n\n#### Advanced\nThe implicit set and prepend operations above cover the most common use cases when modifying environment variables. For more advanced use cases you can use the following explicit operation keys.\n```\nSVAR:\n    set:\n        - Value0\n        - Value1\nPVAR:\n    prepend:\n        - X\nRVAR:\n    unset: 1\nPATH:\n    remove:\n        - C:/Python27\n        - C:/Python27/Scripts\nPYTHONPATH:\n    append:\n        - $MODULE/python\n        - $MODULE/lib\n```\n\nYou can also uses lists of opreations to perform complex modifications.\n```\nPATH:\n    - remove: /some/file/path\n    - append: /some/other/path\n    - prepend: /one/more/path\n```\n\nOne workflow that this enables is the use of modules solely for the purpose of overriding environment variables. Imagine you have a module `my_tool` and it uses a variable `MY_TOOL_PLUGINS` to lookup plugins.\n```\nname: my_tool\n...\nenvironment:\n    MY_TOOL_PLUGINS:\n        - //studio/dev/my_tool/plugins\n        - //studio/projects/test_project/plugins\n```\n\nNow imagine you have a new project and you want `my_tool` to look at a different location for plugins just for that project. Rather than create a new version of the `my_tool` module, create a override module. We might name this module after our project, `project_b`.\n```\nname: project_b\n...\nenvironment:\n    MY_TOOL_PLUGINS:\n        set:\n            - //studio/prod/my_tool/plugins\n            - //studio/projects/project_b/plugins\n```\n\nAll we have to do is activate `my_tool` and `project_b` in that order to make sure our overrides are used.\n```\n> cpenv activate my_tool project_b\n```\n\n#### Requires key\nThe requires key is a list of dependencies that a module needs to function. Currently this is only used for reference, these modules will not be activated automatically.\n\n## Test a Module\nWhen you\'re working on a module navigate into it\'s root directory. Then you can activate it using `cpenv activate .`. This is\nthe best way to validate your module prior to publishing.\n\n## Publish a Module\nOnce you\'re Module is ready for production, use `cpenv publish .` to publish it. Publishing a Module uploads it to a Repo of your choosing.\n\n# Repos\nRepos are storage locations for Modules that support finding, listing, uploading, and downloading Modules via *requirements* like\n`my_module-0.1.0`. Cpenv is configured with the following LocalRepos by default:\n\n- **cwd** - Your current working directory\n- **user** - A user specific repo\n- **home** - A machine wide repo\n\nUse `cpenv repo list` to display your configured Repos. LocalRepos point directly to folders on your local file system.\nCustomize the home Repo by setting the `CPENV_HOME` environment variable.\n\nWhen you activate a module using a requirement, all configured Repos are searched and the best match is used. If the resolved\nmodule is not in a LocalRepo it will be downloaded to your home Repo then activated. This is one of the key features of cpenv\nand allows for strong distributed workflows. For example, you can configure a remote repo like the ShotgunRepo and store your modules directly in a\n[Shotgun studio](https://www.shotgunsoftware.com/) database. [Visit the tk-cpenv repository for more info on using cpenv with Shotgun](https://github.com/cpenv/tk-cpenv)\n\n# Requirements\nRequirements are strings used to resolve and activate modules in Repos. They can be versionless like `my_module` or require a\nversion like `my_module-0.1.0`. Cpenv supports semver/calver, simple versions (v1), and what I like to call *weird* versions\nlike 12.0v2 (The Foundry products). In the future cpenv may support more complex requirements by utilizing\n[resolvelib](https://github.com/sarugaku/resolvelib).\n',
    'author': 'Dan Bradham',
    'author_email': 'danielbradham@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-

# Standard library imports
import os
from datetime import datetime

# Local imports
from . import __version__, http
from .api import get_cache_path
from .versions import default_version, parse_version

_cache_expiration = 3600  # seconds or 1 hour
_warning_template = '''
WARNING: You are using cpenv version {}, however version {} is available.
Please consider upgrading using one of the following commands.

    pipx upgrade cpenv
    pip install --upgrade cpenv
'''


def is_latest_version():
    '''Check if the current version is the latest version.

    Returns:
        (bool, Version, Version) - is_latest, current, latest
    '''

    current = get_current_version()
    latest = get_latest_version()
    return current >= latest, current, latest


def warn_newer_version_available(current, latest):
    '''Warn user that a new version is available.

    To be used in conjunction with is_latest_version.
    '''

    print(_warning_template.format(current.string, latest.string))


def get_current_version():
    '''Return the current version of cpenv as a Version object'''

    return parse_version(__version__)


def get_latest_version():
    '''Use pip search to query pypi for the latest version of cpenv.'''

    latest_file = get_cache_path('latest_version')
    latest_file_expired = True
    if os.path.isfile(latest_file):
        time_since_modified = (
            datetime.now()
            - datetime.fromtimestamp(os.path.getmtime(latest_file))
        ).total_seconds()
        latest_file_expired = time_since_modified > _cache_expiration

    latest = default_version()
    if latest_file_expired:
        try:
            response = http.get('https://pypi.python.org/pypi/cpenv/json')
            json = http.json(response)
            latest = parse_version(json['info']['version'])
            if latest:
                with open(latest_file, 'w') as f:
                    f.write(latest.string)
        except Exception as e:
            print('Failed to query pypi for latest version: %s' % e)

            with open(latest_file, 'w') as f:
                f.write(default_version().string)

    else:
        with open(latest_file, 'r') as f:
            latest_string = f.read()

        latest = parse_version(latest_string)

    return latest



import os

from cpenv import api, paths
from cpenv.cli import core
from cpenv.module import parse_module_path


class Create(core.CLI):
    '''Create a new Module.'''

    def setup_parser(self, parser):
        parser.add_argument(
            'where',
            help='Path to new module',
        )

    def run(self, args):
        where = paths.normalize(args.where)
        if os.path.isdir(where):
            core.echo()
            core.echo('Error: Can not create module in existing directory.')
            core.exit(1)

        default_name, default_version = parse_module_path(where)

        core.echo()
        core.echo('This command will guide you through creating a new module.')
        core.echo()
        name = core.prompt('  Module Name [%s]: ' % default_name)
        version = core.prompt('  Version [%s]: ' % default_version.string)
        description = core.prompt('  Description []: ')
        author = core.prompt('  Author []: ')
        email = core.prompt('  Email []: ')

        core.echo()
        core.echo('- Creating your new Module...', end='')
        module = api.create(
            where=where,
            name=name or default_name,
            version=version or default_version.string,
            description=description,
            author=author,
            email=email,
        )
        core.echo('OK!')

        core.echo()
        core.echo('  ' + module.path)

        core.echo()
        core.echo('Steps you might take before publishing...')
        core.echo()
        core.echo('  - Include binaries your module depends on')
        core.echo('  - Edit the module.yml file')
        core.echo('    - Add variables to the environment section')
        core.echo('    - Add other modules to the requires section')
        core.echo('  - Add python hooks like post_activate')
        core.echo()
